from .maxcut_tfim_util import binary_search, opencl_context, to_scipy_sparse_upper_triangular
from .tsp import tsp_symmetric, tsp_symmetric_sparse, tsp_symmetric_streaming
from .spin_glass_solver import spin_glass_solver
from .spin_glass_solver_sparse import spin_glass_solver_sparse
from .spin_glass_solver_streaming import spin_glass_solver_streaming
import networkx as nx
from numba import njit, prange


@njit(parallel=True)
def tsp_to_maxcut_bipartition(tsp_path, weights):
    n = len(tsp_path)
    best_cut_value = -float('inf')
    best_partition_A = None
    best_partition_B = None
    direction = 0

    for offset in [-1, 0, 1]:
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                cut_value += weights[u, v]
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            direction = offset

    if direction == 0:
        return best_partition_A, best_partition_B, best_cut_value

    improved = True
    best_offset = direction
    while improved:
        improved = False
        offset = best_offset + direction
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                cut_value += weights[u, v]
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            best_offset = offset
            improved = True

    return best_partition_A, best_partition_B, best_cut_value


@njit
def get_cut(bitstring, nodes):
    solution = list(bitstring)
    l, r = [], []
    for i in range(len(solution)):
        if solution[i] == "1":
            r.append(nodes[i])
        else:
            l.append(nodes[i])

    return l, r


@njit
def get_bitstring(partition, nodes):
    bitstring = ""
    for node in nodes:
        bitstring += "0" if node in partition[0] else "1"

    return bitstring


@njit(parallel=True)
def early_exit(G_m, partition, nodes):
    n_qubits = len(nodes)
    bitstring = get_bitstring(partition, nodes)
    theta_bits = [(b == "1") for b in list(bitstring)]
    energy = 0
    for u in prange(n_qubits):
        for v in range(u + 1, n_qubits):
            val = G_m[nodes[u], nodes[v]]
            energy += (val if theta_bits[u] == theta_bits[v] else -val)

    return bitstring, energy


def tsp_maxcut(G, k_neighbors=20, is_optimized=False, is_parallel=True, **kwargs):
    dtype = opencl_context.dtype
    G_m = None
    if isinstance(G, nx.Graph):
        nodes = list(G.nodes())
        n_qubits = len(nodes)
        G_m = nx.to_numpy_array(G, weight='weight', nonedge=0.0, dtype=dtype)
    else:
        n_qubits = len(G)
        nodes = list(range(n_qubits))
        G_m = G

    path, _ = tsp_symmetric(G_m, is_cyclic=False, monte_carlo=True, k_neighbors=k_neighbors, is_parallel=is_parallel)
    l, r, cut_value = tsp_to_maxcut_bipartition(path, G_m)

    if not is_optimized:
        bitstring, energy = early_exit(G_m, (l, r), nodes)

        return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy

    bitint = 0
    for b in partition[0]:
        bitint |= 1 << b

    bitstring, cut_value, cut, energy = spin_glass_solver(G_m, best_guess=bitint, **kwargs)
    l, r = get_cut(bitstring, nodes)

    return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy


@njit(parallel=True)
def tsp_to_maxcut_bipartition_sparse(tsp_path, G_data, G_rows, G_cols):
    n = len(tsp_path)
    best_cut_value = -float('inf')
    best_partition_A = None
    best_partition_B = None
    direction = 0

    for offset in [-1, 0, 1]:
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                low, high = (u, v) if u < v else (v, u)
                start = G_rows[low]
                end = G_rows[low + 1]
                i = binary_search(G_cols[start:end], high) + start
                if i < end:
                    cut_value += G_data[i]
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            direction = offset

    if direction == 0:
        return best_partition_A, best_partition_B, best_cut_value

    improved = True
    best_offset = direction
    while improved:
        improved = False
        offset = best_offset + direction
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                low, high = (u, v) if u < v else (v, u)
                start = G_rows[low]
                end = G_rows[low + 1]
                i = binary_search(G_cols[start:end], high) + start
                if i < end:
                    cut_value += G_data[i]
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            best_offset = offset
            improved = True

    return best_partition_A, best_partition_B, best_cut_value


@njit(parallel=True)
def early_exit_sparse(G_data, G_rows, G_cols, partition, nodes):
    n_qubits = len(nodes)
    bitstring = get_bitstring(partition, nodes)
    theta_bits = [(b == "1") for b in list(bitstring)]
    energy = 0
    for u in prange(n_qubits):
        for v in range(u + 1, n_qubits):
            start = G_rows[u]
            end = G_rows[u + 1]
            i = binary_search(G_cols[start:end], v) + start
            if i < end:
                val = G_data[i]
                energy += (val if theta_bits[u] == theta_bits[v] else -val)

    return bitstring, energy


def tsp_maxcut_sparse(G, k_neighbors=20, is_optimized=False, is_parallel=True, **kwargs):
    dtype = opencl_context.dtype
    G_m = None
    if isinstance(G, nx.Graph):
        nodes = list(G.nodes())
        n_qubits = len(nodes)
        G_m = to_scipy_sparse_upper_triangular(G, nodes, n_qubits, dtype)
    else:
        n_qubits = G.shape[0]
        nodes = list(range(n_qubits))
        G_m = G

    path, _ = tsp_symmetric_sparse(G_m, k_neighbors=k_neighbors, is_parallel=is_parallel)
    l, r, cut_value = tsp_to_maxcut_bipartition_sparse(path, G_m.data, G_m.indptr, G_m.indices)

    if not is_optimized:
        bitstring, energy = early_exit_sparse(G_m.data, G_m.indptr, G_m.indices, (l, r), nodes)

        return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy

    bitint = 0
    for b in partition[0]:
        bitint |= 1 << b

    bitstring, cut_value, cut, energy = spin_glass_solver_sparse(G_m, best_guess=bitint, **kwargs)
    l, r = get_cut(bitstring, nodes)

    return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy


@njit(parallel=True)
def tsp_to_maxcut_bipartition_streaming(tsp_path, G_func):
    n = len(tsp_path)
    best_cut_value = -float('inf')
    best_partition_A = None
    best_partition_B = None
    direction = 0

    for offset in [-1, 0, 1]:
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                cut_value += G_func(u, v)
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            direction = offset

    if direction == 0:
        return best_partition_A, best_partition_B, best_cut_value

    improved = True
    best_offset = direction
    while improved:
        improved = False
        offset = best_offset + direction
        mid = n // 2 + offset
        A = tsp_path[:mid]
        B = tsp_path[mid:]
        cut_value = 0.0
        for i in prange(len(A)):
            u = A[i]
            for v in B:
                cut_value += G_func(u, v)
        if cut_value > best_cut_value:
            best_cut_value = cut_value
            best_partition_A = A
            best_partition_B = B
            best_offset = offset
            improved = True

    return best_partition_A, best_partition_B, best_cut_value


@njit(parallel=True)
def early_exit_streaming(G_func, partition, nodes):
    n_qubits = len(nodes)
    bitstring = get_bitstring(partition, nodes)
    theta_bits = [(b == "1") for b in list(bitstring)]
    energy = 0
    for u in prange(n_qubits):
        for v in range(u + 1, n_qubits):
            val = G_func(nodes[u], nodes[v])
            energy += (val if theta_bits[u] == theta_bits[v] else -val)

    return bitstring, energy


def tsp_maxcut_streaming(G_func, nodes, k_neighbors=20, is_optimized=False, is_parallel=True, **kwargs):
    n_qubits = len(nodes)
    path, _ = tsp_symmetric_streaming(G_func, nodes, k_neighbors=k_neighbors, is_parallel=is_parallel)
    l, r, cut_value = tsp_to_maxcut_bipartition_streaming(path, G_func)

    if not is_optimized:
        bitstring, energy = early_exit_streaming(G_func, (l, r), nodes)

        return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy

    bitint = 0
    for b in partition[0]:
        bitint |= 1 << b

    bitstring, cut_value, cut, energy = spin_glass_solver_streaming(G_func, nodes, best_guess=bitint, **kwargs)
    l, r = get_cut(bitstring, nodes)

    return bitstring, cut_value, ([nodes[x] for x in l], [nodes[x] for x in r]), energy



import math
import networkx as nx
import numpy as np
import os
from numba import njit, prange

from .maxcut_tfim_util import fix_cdf, get_cut, init_theta, init_thresholds, maxcut_hamming_cdf, opencl_context


epsilon = opencl_context.epsilon


@njit
def update_repulsion_choice(G_m, max_edge, weights, n, used, node):
    # Select node
    used[node] = True

    # Repulsion: penalize neighbors
    for nbr in range(n):
        if used[nbr]:
            continue
        weights[nbr] *= max(epsilon, 1 - G_m[node, nbr] / max_edge)


# Written by Elara (OpenAI custom GPT) and improved by Dan Strano
@njit
def local_repulsion_choice(G_m, max_edge, weights, tot_init_weight, n, m):
    """
    Pick m nodes out of n with repulsion bias:
    - High-degree nodes are already less likely
    - After choosing a node, its neighbors' probabilities are further reduced
    adjacency_data, adjacency_rows: CSR-format sparse adjacency data
    weights: float32 array of shape (n,)
    """

    weights = weights.copy()
    used = np.zeros(n, dtype=np.bool_) # False = available, True = used

    # First bit:
    r = np.random.rand()
    cum = 0.0
    node = 0
    for i in range(n):
        cum += weights[i]
        if (tot_init_weight * r) < cum:
            node = i
            break

    if m == 1:
        used[node] = True
        return used

    update_repulsion_choice(G_m, max_edge, weights, n, used, node)

    for _ in range(1, m - 1):
        # Count available
        total_w = 0.0
        for i in range(n):
            if used[i]:
                continue
            total_w += weights[i]

        # Normalize & sample
        r = np.random.rand()
        cum = 0.0
        node = -1
        for i in range(n):
            if used[i]:
                continue
            cum += weights[i]
            if (total_w * r) < cum:
                node = i
                break

        if node == -1:
            node = 0
            while used[node]:
                node += 1

        # Update answer and weights
        update_repulsion_choice(G_m, max_edge, weights, n, used, node)

    # Count available
    total_w = 0.0
    for i in range(n):
        if used[i]:
            continue
        total_w += weights[i]

    # Normalize & sample
    r = np.random.rand()
    cum = 0.0
    node = -1
    for i in range(n):
        if used[i]:
            continue
        cum += weights[i]
        if (total_w * r) < cum:
            node = i
            break

    if node == -1:
        node = 0
        while used[node]:
            node += 1

    used[node] = True

    return used


@njit
def compute_energy(sample, G_m, n_qubits):
    energy = 0
    for u in range(n_qubits):
        for v in range(u + 1, n_qubits):
            val = G_m[u, v]
            energy += val if sample[u] == sample[v] else -val

    return energy


@njit(parallel=True)
def sample_for_solution(G_m, shots, thresholds, weights, dtype):
    shots = max(1, shots >> 1)
    n = len(G_m)
    max_edge = G_m.max()
    tot_init_weight = weights.sum()

    solutions = np.empty((shots, n), dtype=np.bool_)
    energies = np.empty(shots, dtype=dtype)

    best_solution = solutions[0]
    best_energy = float("inf")

    improved = True
    while improved:
        improved = False
        for s in prange(shots):
            # First dimension: Hamming weight
            mag_prob = np.random.random()
            m = 0
            while thresholds[m] < mag_prob:
                m += 1
            m += 1

            # Second dimension: permutation within Hamming weight
            sample = local_repulsion_choice(G_m, max_edge, weights, tot_init_weight, n, m)
            solutions[s] = sample
            energies[s] = compute_energy(sample, G_m, n)

        best_index = np.argmin(energies)
        energy = energies[best_index]
        if energy < best_energy:
            best_energy = energy
            best_solution = solutions[best_index].copy()
            improved = True

    best_value = 0.0
    for u in range(n):
        for v in range(u + 1, n):
            if best_solution[u] != best_solution[v]:
                best_value += G_m[u, v]

    return best_solution, best_value


@njit(parallel=True)
def init_J_and_z(G_m, dtype):
    n_qubits = len(G_m)
    degrees = np.empty(n_qubits, dtype=np.uint32)
    J_eff = np.empty(n_qubits, dtype=dtype)
    J_max = -float("inf")
    for n in prange(n_qubits):
        degree = sum(G_m[n] != 0.0)
        J = (-G_m[n].sum() / degree) if degree > 0 else 0
        degrees[n] = degree
        J_eff[n] = J
        J_abs = abs(J)
        J_max = max(J_abs, J_max)
    J_eff /= J_max

    return J_eff, degrees


@njit
def cpu_footer(shots, quality, n_qubits, G_m, nodes, dtype):
    J_eff, degrees = init_J_and_z(G_m, dtype)
    hamming_prob = init_thresholds(n_qubits, dtype)

    maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, hamming_prob, dtype)

    degrees = None
    J_eff = 1.0 / (1.0 + epsilon - J_eff)

    best_solution, best_value = sample_for_solution(G_m, shots, hamming_prob, J_eff, dtype)

    bit_string, l, r = get_cut(best_solution, nodes)

    return bit_string, best_value, (l, r)


def maxcut_tfim(
    G,
    quality=None,
    shots=None
):
    dtype = opencl_context.dtype
    wgs = opencl_context.work_group_size
    nodes = None
    n_qubits = 0
    G_m = None
    if isinstance(G, nx.Graph):
        nodes = list(G.nodes())
        n_qubits = len(nodes)
        G_m = nx.to_numpy_array(G, weight='weight', nonedge=0.0, dtype=dtype)
    else:
        n_qubits = len(G)
        nodes = list(range(n_qubits))
        G_m = G

    if n_qubits < 3:
        if n_qubits == 0:
            return "", 0, ([], [])

        if n_qubits == 1:
            return "0", 0, (nodes, [])

        if n_qubits == 2:
            weight = G_m[0, 1]
            if weight < 0.0:
                return "00", 0, (nodes, [])

            return "01", weight, ([nodes[0]], [nodes[1]])

    is_segmented = (G_m.nbytes << 1) > opencl_context.max_alloc
    if is_segmented and is_alt_gpu_sampling:
        print("[WARN] Using segmented solver, so disabling is_alt_gpu_sampling.")
        is_alt_gpu_sampling = False

    if quality is None:
        quality = 2

    if shots is None:
        # Number of measurement shots
        shots = n_qubits << quality

    n_steps = n_qubits << quality
    grid_size = n_steps * n_qubits

    return cpu_footer(shots, quality, n_qubits, G_m, nodes, dtype)
